import numpy as np
import pandas as pd

__all__ = [
    "rebin", "merge",
    "analysis_aggcols",
    "analysis_pivot_merge",
    "analysis_pivot_variations",
]

def rebin(df, bins, label="binvar0", underflow=True, overflow=True):
    """
    Take a dataframe with the 'label' as an index and rebin/densify into the
    input bins returning a new dataframe.

    Parameters
    ----------
    df : DataFrame-like
        Dataframe with index `label` which will be rebinned (and densified)
        into a new set of bins.

    bins : ndarray-like
        The bins to rebin into.

    label : str (default, binvar0)
        The index label to rebin. Other indexes are left alone.

    underflow : bool (default, True)
        Include underflow entires in the lowest bin

    overflow : bool (default, True)
        Include overflow entries in the highest bin
    """
    index_names = df.index.names
    tdf = df.reset_index()
    #index_arr = bins[np.minimum(bins.searchsorted(tdf[label]), len(bins)-1)]
    tidx = bins.searchsorted(tdf[label], side='right')
    if underflow:
        tidx = np.maximum(tidx, 1)
    if overflow:
        tidx = np.minimum(tidx, len(bins))
    index_arr = bins[tidx-1]
    tdf.loc[:,label] = index_arr
    tdf = tdf.groupby(index_names).sum()
    index_names = [idx for idx in tdf.index.names if idx!=label]
    if len(index_names)==0:
        ret = tdf.reindex(bins).fillna(0.).groupby(label).sum()
    else:
        ret = (
            tdf.groupby(index_names)
            .apply(
                lambda dfgrp: (
                    dfgrp.reset_index(index_names, drop=True)
                    .reindex(bins).fillna(0.)
                )
            )
        )
    return ret

def merge(df, merge_cfg, index_level="parent"):
    """
    Change certain names in an index_level and merge

    Parameters
    ----------
    df : DataFrame-like
        Dataframe input to change some index levels

    merge_cfg : dict
        Dictionary of new labels with values corresponding to a list of
        index values to change into the new label

    index_level : str (default, parent)
        Index to change
    """
    index_names = df.index.names
    tdf = df.reset_index()
    for label, match_labels in merge_cfg.items():
        tdf.loc[tdf[index_level].isin(match_labels), index_level] = label
    return tdf.groupby(index_names).sum()

def merge_query(df, label_queries, index_level="parent"):
    index_names = df.index.names
    tdf = df.reset_index()
    for label, query in label_queries.items():
        mask = tdf.eval(query)
        tdf.loc[mask, index_level] = label
    return tdf.groupby(index_names).sum()

def analysis_aggcols(
    df, variation="lhePdfWeight", nominal="sum_w_{}", variance="sum_ww_{}",
    aggs=["std"], evals=[
        "sum_w_lhePdfWeightUp = sum_w + agg_std",
        "sum_w_lhePdfWeightDown = sum_w - agg_std",
        "sum_ww_lhePdfWeightUp = sum_ww * (1 + agg_std/sum_w)**2",
        "sum_ww_lhePdfWeightDown = sum_ww * (1 - agg_std/sum_w)**2",
    ], copy=True,
):
    """
    Aggregate across a set of columns and apply some function

    Parameters
    ----------
    df : DataFrame-like
        Dataframe input to process

    variation : str (default, "lhePdfWeight")
        Variation which appears in the columns

    nominal : str (default, "sum_w_{}")
        How the nominal weight column is defined

    variance: str (default, "sum_ww_{}")
        How the variance weight column is defined

    aggs : list (default, ["std"])
        List of aggregation functions to apply. Will add columns for these

    evals : list (default, [...])
        List of evals to generate new columns

    copy : bool (default, True)
        Copy the input dataframe
    """
    columns = [c for c in df.columns if nominal.format(variation) in c]

    tdf = df.copy(deep=True) if copy else df
    for agg in aggs:
        tdf["agg_{}".format(agg)] = tdf.loc[:, columns].agg(agg, axis=1)

    columns.extend([c for c in tdf.columns if variance.format(variation) in c])
    tdf = tdf.drop(columns, axis=1)

    tdf.eval("\n".join(evals), inplace=True)
    return tdf.drop(["agg_{}".format(agg) for agg in aggs], axis=1)

def analysis_pivot_merge(df, df_alt, label='table', new_value="Event"):
    columns = df_alt.columns
    index = df_alt.index.names

    df_piv = pd.pivot_table(
        df_alt, values=columns, index=[c for c in index if c!=label],
        columns=[label], fill_value=0.,
    )
    df_piv.columns = ["_".join(c) for c in df_piv.columns]
    df_piv[label] = new_value
    df_piv = df_piv.set_index("table", append=True)
    df_piv = df_piv.reorder_levels(index)
    return pd.concat([df, df_piv], axis='columns')

def analysis_pivot_variations(df, label="variation"):
    columns = df.columns
    index = df.index.names

    # Change column names to column values
    df.columns.name = label
    df = df.stack().reset_index(label)

    # Add quantity for sum_w/sum_ww and remove from variation
    df["quantity"] = "sum_w"
    df.loc[df[label].str.startswith("sum_ww"), "quantity"] = "sum_ww"
    df.loc[:,label] = (
        df[label]
        .str.replace("sum_ww_", "")
        .str.replace("sum_w_", "")
        .str.replace("sum_ww", "")
        .str.replace("sum_w", "")
    )

    # Make quantity column values in column labels and label into an index
    df = df.set_index([label, "quantity"], append=True).unstack()
    df.columns = [c[1] for c in df.columns]
    return df

def analysis_flat_variations(df, label="variation"):
    def check_inputs(data):
        bad_cols = not all(c in data.columns for c in ["sum_w", "sum_ww"])
        bad_index = not (list(data.index.names[-3:]) == [label, "bin_min", "bin_max"])
        bad_dups = data.index.duplicated().any()

        if bad_cols:
            raise KeyError(
                "('sum_w', 'sum_ww') need to be in the columns. Found "
                "{}".format(data.columns)
            )
        if bad_index:
            raise IndexError(
                "('{}', 'bin_min', 'bin_max') should appear as the last 3 "
                "indexes. Found {}".format(label, data.index.names)
            )
        if bad_dups:
            raise ValueError(
                "Found duplicate indexes: "
                "{}".format(df.loc[df.index.duplicated(),:])
            )
    check_inputs(df)

    # ["region", "process", "variation", "bin_min", "bin_max"]
    index = df.index.names
    idx_variation = index.index(label)

    mask_var = (df.index.get_level_values(label)=="")
    dfn = df.loc[mask_var,:].copy()
    dfv = df.loc[~mask_var,:].copy()

    dfn_sum = dfn.groupby(index[:idx_variation]).sum()
    dfv_sum = dfv.groupby(index[:idx_variation+1]).sum()
    dfv_sum = dfv_sum.divide(dfn_sum)
    dfv_sum.columns = ["scale_w", "scale_ww"]

    dfn = dfn.reset_index(label, drop=True)

    dfv = pd.merge(
        dfv_sum["scale_w"],
        dfn[["sum_w", "sum_ww"]],
        right_index=True, left_index=True,
    ).reorder_levels(index).sort_index()
    dfv.loc[:,"sum_w"] = dfv.eval("scale_w*sum_w")
    dfv.loc[:,"sum_ww"] = dfv.eval("(scale_w**2)*sum_ww")
    dfv = dfv.drop("scale_w", axis=1)

    dfn[label] = ""
    dfn = dfn.set_index(label, append=True).reorder_levels(index).sort_index()
    return pd.concat([dfn, dfv], axis=0, sort=True)

def analysis_smooth_variations(df, smoother, label="variation"):
    def check_inputs(data):
        bad_cols = not all(c in data.columns for c in ["sum_w", "sum_ww"])
        bad_index = not (list(data.index.names[-3:]) == [label, "bin_min", "bin_max"])
        bad_dups = data.index.duplicated().any()

        if bad_cols:
            raise KeyError(
                "('sum_w', 'sum_ww') need to be in the columns. Found "
                "{}".format(data.columns)
            )
        if bad_index:
            raise IndexError(
                "('{}', 'bin_min', 'bin_max') should appear as the last 3 "
                "indexes. Found {}".format(label, data.index.names)
            )
        if bad_dups:
            raise ValueError(
                "Found duplicate indexes: "
                "{}".format(df.loc[df.index.duplicated(),:])
            )
    check_inputs(df)

    # ["region", "process", "variation", "bin_min", "bin_max"]
    index = df.index.names
    idx_variation = index.index(label)

    mask_var = (df.index.get_level_values(label)=="")
    dfv = df[~mask_var].reset_index(index[idx_variation:])
    dfv["updown"] = "Up"
    dfv.loc[dfv[label].str.endswith("Down"), "updown"] = "Down"
    dfv.loc[:, label] = np.where(
        dfv[label].str.endswith("Down"),
        dfv[label].str.slice(0, -4),
        dfv[label].str.slice(0, -2),
    )
    dfv = (
        dfv.set_index(list(index[idx_variation:])+["updown"], append=True)
        .unstack()
    )
    dfv.columns = ["_".join(c) for c in dfv.columns]

    # add nominal
    dfv = pd.merge(
        dfv, df.loc[mask_var,:].reset_index(label, drop=True),
        right_index=True, left_index=True,
    )
    dfv["bin_cent"] = 0.5*(
        dfv.index.get_level_values(index[-2])
        + dfv.index.get_level_values(index[-1])
    )
    dfv["up_norm"] = dfv.eval("sum_w_Up/sum_w")
    dfv["do_norm"] = dfv.eval("sum_w_Down/sum_w")
    dfv["up_err"] = dfv.eval("sqrt(abs(sum_ww_Up - sum_ww_Down))/sum_w")
    dfv["do_err"] = dfv.eval("sqrt(abs(sum_ww_Down - sum_ww))/sum_w")

    # swap values
    mask_nonzero = (dfv["do_err"]==0.)
    dfv.loc[mask_nonzero, "do_err"] = dfv.loc[mask_nonzero, "up_err"].copy()
    dfv.loc[:, "up_err"] = np.maximum(1e-4, np.minimum(1., dfv["up_err"]))
    dfv.loc[:, "do_err"] = np.maximum(1e-4, np.minimum(1., dfv["do_err"]))

    # smooth
    dfv = pd.concat([
        dfv, dfv.groupby(index[:idx_variation+1]).apply(smoother),
    ], axis=1)

    # normalise
    dfv["sum_w_upsmooth"] = dfv.eval("sum_w*upsmooth")
    dfv["sum_w_dosmooth"] = dfv.eval("sum_w*dosmooth")
    dfv_sum = dfv[[
        "sum_w", "sum_w_Up", "sum_w_Down", "sum_w_upsmooth", "sum_w_dosmooth",
    ]].groupby(index[:idx_variation+1]).sum()
    dfv = dfv.reorder_levels(index).sort_index()

    sum_w_up = dfv.eval("sum_w*upsmooth").mul(dfv_sum.eval("sum_w_Up/sum_w_upsmooth"))
    sum_w_do = dfv.eval("sum_w*dosmooth").mul(dfv_sum.eval("sum_w_Down/sum_w_dosmooth"))

    dfv = pd.DataFrame({
        "sum_wUp": sum_w_up.values,
        "sum_wwUp": dfv["sum_ww_Up"].values,
        "sum_wDown": sum_w_do.values,
        "sum_wwDown": dfv["sum_ww_Down"].values,
    }, index=dfv.index)

    dfv.columns.name = "type"
    dfv = dfv.stack().reset_index([label, "type"])
    dfv.loc[:, label] = np.where(
        dfv["type"].str.endswith("Up"), dfv[label]+"Up", dfv[label]+"Down",
    )
    dfv.loc[:, "type"] = np.where(
        dfv["type"].str.endswith("Up"),
        dfv["type"].str.slice(0, -2),
        dfv["type"].str.slice(0, -4),
    )
    dfv = (
        dfv.set_index([label, "type"], append=True)
        .unstack().reorder_levels(index).sort_index()
    )
    dfv.columns = [c[1] for c in dfv.columns]

    return dfv



import pysge
import copy
import numpy as np
import pandas as pd
import scipy.stats
import numdifftools
import iminuit

from .stats import poisson_interval
from .draw import process_colours, process_names

class NLLModel(object):
    def __init__(
        self, data, mc, config, same_bin_widths=False,
        shape_type="shape", smooth_region=1.,
        observable_function = "bin_min, bin_max: (bin_min + bin_max)/2."
    ):
        self.verbose = False
        self.dfdata = data
        self.dfmc = mc
        self.config = config

        self.shape_type = shape_type
        self.smooth_region = smooth_region

        self.asimov = False
        self.toy = -1
        self.seed = 123456
        self.saturated = False

        self.norm_nuisances = [
            p["name"]
            for p in config["parameters"]
            if p["constraint"] == "gaussian"
        ]
        self.morph_nuisances = [
            p["name"]
            for p in config["parameters"]
            if p["constraint"] == "gaussian"
        ]

        region_procs = config["regions"]
        self.region_procs = region_procs
        self.regions = list(region_procs.keys())

        self.mcstat_nuisances = {
            r: [p["name"] for p in config["parameters"] if r in p["name"] and p["constraint"]=="gamma"]
            for r in self.regions
        }

        self.param_names = []
        self.param_inits = []
        self.param_fixes = []
        self.param_limit = []
        for idx, params in enumerate(config["parameters"]):
            self.param_names.append(params["name"])
            self.param_inits.append(params["value"])
            self.param_fixes.append(params["fixed"])
            self.param_limit.append(params["limit"])

        self.bins = {}
        self.data = {}
        self.obs = {}
        self.dbins = {}
        for r in self.regions:
            self.data[r] = data.loc[(r,),"sum_w"].values
            bins = data.loc[(r,),:].reset_index()[["bin_min", "bin_max"]].values
            self.obs[r] = eval("lambda "+observable_function)(bins[:,0], bins[:,1])
            self.dbins[r] = bins[:,1] - bins[:,0]
            if same_bin_widths:
                self.dbins[r][:] = 1.
            self.bins[r] = (bins[:,0], bins[:,1])

        # self.data_asimov[r,b]
        # self.data_toy[r,b]

        self.scale_functions = {}
        for r in self.regions:
            self.scale_functions[r] = []
            functions = []
            for p in region_procs[r]:
                self.scale_functions[r].append(
                    config["scale_functions"].get((r, p), "x, w, p: 1.")
                )

        self.sumw = {}
        self.sumww = {}
        for r in self.regions:
            self.sumw[r] = []
            self.sumww[r] = []
            for p in region_procs[r]:
                self.sumw[r].append(mc.loc[(r,p,""),"sum_w"].values)
                self.sumww[r].append(mc.loc[(r,p,""),"sum_ww"].values)
            self.sumw[r] = np.array(self.sumw[r])
            self.sumww[r] = np.array(self.sumww[r])

        self.nom_sum = {}
        self.up_sum = {}
        self.do_sum = {}
        for r in self.regions:
            self.nom_sum[r] = []
            self.up_sum[r] = []
            self.do_sum[r] = []
            for p in region_procs[r]:
                self.nom_sum[r].append(mc.loc[(r,p,""),"sum_w"].sum())

                self.up_sum[r].append([])
                self.do_sum[r].append([])
                for n in self.norm_nuisances:
                    try:
                        self.up_sum[r][-1].append(
                            mc.loc[(r,p,n+"Up"),"sum_w"].sum()
                        )
                        self.do_sum[r][-1].append(
                            mc.loc[(r,p,n+"Down"),"sum_w"].sum()
                        )
                    except KeyError:
                        self.up_sum[r][-1].append(self.nom_sum[r][-1])
                        self.do_sum[r][-1].append(self.nom_sum[r][-1])
            self.nom_sum[r] = np.array(self.nom_sum[r])
            self.up_sum[r] = np.array(self.up_sum[r])
            self.do_sum[r] = np.array(self.do_sum[r])

        self.delta_sum = {}
        self.delta_diff = {}
        for r in self.regions:
            self.delta_sum[r] = []
            self.delta_diff[r] = []
            for p in region_procs[r]:
                self.delta_sum[r].append([])
                self.delta_diff[r].append([])

                nom = mc.loc[(r, p, ""), "sum_w"].values
                nom_norm = nom/(nom*self.dbins[r]).sum()
                for n in self.morph_nuisances:
                    skip = False
                    try:
                        up = mc.loc[(r,p,n+"Up"),"sum_w"].values
                        do = mc.loc[(r,p,n+"Down"),"sum_w"].values
                    except KeyError:
                        delta_sum = np.zeros_like(nom_norm)
                        delta_diff = np.zeros_like(nom_norm)
                        skip = True

                    if not skip:
                        up_norm = up/(up*self.dbins[r]).sum()
                        do_norm = do/(do*self.dbins[r]).sum()

                        if self.shape_type == "shape":
                            up_del = up_norm - nom_norm
                            do_del = do_norm - nom_norm
                        elif self.shape_type == "shapeN":
                            up_del = np.where(
                                (up_norm>0)&(nom_norm>0),
                                np.log(up_norm/nom_norm),
                                np.zeros_like(nom_norm),
                            )
                            do_del = np.where(
                                (do_norm>0)&(nom_norm>0),
                                np.log(do_norm/nom_norm),
                                np.zeros_like(nom_norm),
                            )
                        delta_diff = up_del - do_del
                        delta_sum = up_del + do_del
                    self.delta_diff[r][-1].append(delta_diff)
                    self.delta_sum[r][-1].append(delta_sum)
            self.delta_diff[r] = np.swapaxes(np.array(self.delta_diff[r]), 1, 2)
            self.delta_sum[r] = np.swapaxes(np.array(self.delta_sum[r]), 1, 2)

        self.sumw_procsum = {
            r: (
                mc.loc[(r, pd.IndexSlice[:], ""), "sum_w"]
                .groupby(["bin_min", "bin_max"]).sum().values
            ) for r in self.regions
        }
        self.sumww_procsum = {
            r: (
                mc.loc[(r, pd.IndexSlice[:], ""), "sum_ww"]
                .groupby(["bin_min", "bin_max"]).sum().values
            ) for r in self.regions
        }
        self.neff_procsum = {
            r: self.sumw_procsum[r]**2/self.sumww_procsum[r]
            for r in self.regions
        }

        self.parameters = {}
        self.set_parameters({p["name"]: p["value"] for p in config["parameters"]})

    def _shape_norm(self, region):
        norm = np.ones_like(self.up_sum[region])
        condition = (self.theta_norm>=0.)
        norm[:,condition] = (self.up_sum[region]/self.nom_sum[region][:,np.newaxis])[:,condition]
        norm[:,~condition] = (self.do_sum[region]/self.nom_sum[region][:,np.newaxis])[:,~condition]
        val = np.power(norm, np.abs(self.theta_norm[np.newaxis,:]))
        val[np.isinf(val)] = 0.
        return val

    def shape_norm(self, region):
        df = pd.DataFrame(
            self._shape_norm(region),
            columns=self.norm_nuisances,
            index=self.region_procs[region],
        ).stack()
        df.index.names = ["process", "nuisance"]
        df.columns = ["norm"]
        return df

    def _shape_norms(self, region):
        val = self._shape_norm(region)
        if np.isnan(val).any():
            raise RuntimeError("NaN in shape_norm({}): {}".format(region, val))
        if np.isinf(val).any():
            raise RuntimeError("Inf in shape_norm({}): {}".format(region, val))
        return np.prod(self._shape_norm(region), axis=1)

    def shape_norms(self, region):
        df = pd.DataFrame(
            self._shape_norms(region),
            columns=["norm"],
            index=self.region_procs[region],
        )
        df.index.names = ["process"]
        return df

    def _shape_morph(self, region):
        theta_norm = self.theta_morph/self.smooth_region
        theta_norm2 = (theta_norm)**2

        # smooth step function
        step = np.sign(theta_norm)
        condition = (np.abs(theta_norm)<1.)
        step[condition] = (0.125*theta_norm*(theta_norm2*(3.*theta_norm2-10.)+15.))[condition]

        # theta * (diff + sum*smooth_step_function)
        _theta = self.theta_morph[np.newaxis,np.newaxis,:]
        step = step[np.newaxis,np.newaxis,:]
        return 0.5*_theta*(self.delta_diff[region] + self.delta_sum[region]*step)

    def shape_morph(self, region):
        results = self._shape_morph(region)

        df = pd.DataFrame()
        for i1 in range(results.shape[0]):
            for i2 in range(results.shape[2]):
                tdf = pd.DataFrame(results[i1,:,i2], columns=["morph"])
                tdf["process"] = self.region_procs[region][i1]
                tdf["nuisance"] = self.morph_nuisances[i2]
                tdf["bin_min"] = self.bins[region][0]
                tdf["bin_max"] = self.bins[region][1]
                tdf = tdf.set_index(["process", "nuisance", "bin_min", "bin_max"])
                df = pd.concat([df, tdf], axis=0)
        return df

    def _shape_morphs(self, region):
        sumw_norm = np.sum(self.sumw[region]*self.dbins[region][np.newaxis,:], axis=1)[:,np.newaxis]
        shape_vals_norm = self.sumw[region]/sumw_norm
        if self.shape_type=="shapeN":
            shape_vals_norm = np.log(shape_vals_norm)

        shape_vals_norm += np.sum(self._shape_morph(region), axis=2)
        if np.isnan(shape_vals_norm).any():
            raise RuntimeError("NaN in shape_morph({}): {}".format(region, shape_vals_norm))
        if np.isinf(shape_vals_norm).any():
            raise RuntimeError("Inf in shape_morph({}): {}".format(region, shape_vals_norm))

        if self.shape_type=="shapeN":
            shape_vals_norm = np.exp(shape_vals_norm)
        shape_vals_norm = np.maximum(1e-10, shape_vals_norm)

        return shape_vals_norm*sumw_norm #/sumw

    def shape_morphs(self, region):
        results = self._shape_morphs(region)/self.sumw[region]
        df = pd.DataFrame(results.T, columns=self.region_procs[region])
        df["bin_min"] = self.bins[region][0]
        df["bin_max"] = self.bins[region][1]
        df = df.set_index(["bin_min", "bin_max"]).stack()
        df.index.names = ["bin_min", "bin_max", "process"]
        df.columns = ["morph"]
        df = df.reorder_levels(["process", "bin_min", "bin_max"]).sort_index()
        return df

    def _prediction(self, region):
        nom_scale = np.ones_like(self.sumw[region])
        for idx, function in enumerate(self.scale_functions[region]):
            nom_scale[idx,:] = eval('lambda '+function)(
                self.obs[region], self.sumw[region][idx,:], self.parameters,
            )

        syst_scale = self._shape_morphs(region)*(self._shape_norms(region)[:,np.newaxis])

        #mcstat_scale = (1 + np.sqrt(self.sumww)/self.sumw)**(self.theta_mcstat[:,np.newaxis,:])
        mcstat_scale = np.power(
            (1 + np.sqrt(self.sumww_procsum[region])/self.sumw_procsum[region]),
            self.theta_mcstat[region]
        )[np.newaxis,:]

        # return sumw*nom_scale*syst_scale*mcstat_scale
        # syst_scale already includes sumw
        return nom_scale*syst_scale*mcstat_scale

    def prediction(self, region):
        results = self._prediction(region)
        df = pd.DataFrame(results.T, columns=self.region_procs[region])
        df["bin_min"] = self.bins[region][0]
        df["bin_max"] = self.bins[region][1]
        df = df.set_index(["bin_min", "bin_max"]).stack()
        df.index.names = ["bin_min", "bin_max", "process"]
        df.columns = ["prediction"]
        df = df.reorder_levels(["process", "bin_min", "bin_max"]).sort_index()
        return df

    def _predictions(self, region):
        val = self._prediction(region)
        if np.isnan(val).any():
            raise RuntimeError("NaN in prediction({}): {}".format(region, val))
        if np.isinf(val).any():
            raise RuntimeError("Inf in prediction({}): {}".format(region, val))
        return np.sum(self._prediction(region), axis=0)

    def predictions(self, region):
        results = self._predictions(region)
        df = pd.DataFrame(results, columns=["prediction"])
        df["bin_min"] = self.bins[region][0]
        df["bin_max"] = self.bins[region][1]
        df = df.set_index(["bin_min", "bin_max"])
        return df

    def poisson_pdfs(self, region):
        pred = np.maximum(1e-10, self._predictions(region))

        if self.asimov:
            if self.saturated:
                pred = self.data_asimov[region]
            pdfs = scipy.stats.gamma.logpdf(pred, self.data_asimov[region]+1.)
        elif self.toy>=0:
            if self.saturated:
                pred = self.data_toy[region]
            pdfs = scipy.stats.poisson.logpmf(self.data_toy[region], pred)
        else:
            if self.saturated:
                pred = self.data[region]
            pdfs = scipy.stats.poisson.logpmf(self.data[region], pred)
        if np.isnan(pdfs).any():
            raise RuntimeError("NaN in poisson_pdfs({}): {}".format(region, pdfs))
        if np.isinf(pdfs).any():
            raise RuntimeError("Inf in poisson_pdfs({}): {}".format(region, pdfs))
        return np.sum(pdfs)

    def gaussian_pdfs(self):
        pdfs = scipy.stats.norm.logpdf(self.theta_gaus)
        if np.isnan(pdfs).any():
            raise RuntimeError("NaN in gaussian_pdfs({}): {}".format(region, pdfs))
        if np.isinf(pdfs).any():
            raise RuntimeError("Inf in gaussian_pdfs({}): {}".format(region, pdfs))
        return np.sum(pdfs)

    def gamma_pdfs(self, region):
        sumw = self.sumw_procsum[region]
        sumww = self.sumww_procsum[region]
        neff = self.neff_procsum[region]
        theta = self.theta_mcstat[region]

        scale = np.power((1 + np.sqrt(sumww)/sumw), theta)
        pdfs = scipy.stats.gamma.logpdf(scale*neff, neff+1)
        pdfs[sumww==0.] = 0.
        if np.isnan(pdfs).any():
            raise RuntimeError("NaN in gamma_pdfs({}): {}".format(region, pdfs))
        if np.isinf(pdfs).any():
            raise RuntimeError("Inf in gamma_pdfs({}): {}".format(region, pdfs))
        return np.sum(pdfs)

    def nll(self):
        pois = sum(self.poisson_pdfs(r) for r in self.regions)
        gaus = self.gaussian_pdfs()
        gamm = sum(self.gamma_pdfs(r) for r in self.regions)
        val = -2*(pois+gaus+gamm)
        if self.verbose:
            print("pois: {}\ngaus: {}\ngamm: {}\nnll: {}".format(pois, gaus, gamm, val))
        if np.isnan(val):
            raise RuntimeError("NLL is nan")
        if np.isinf(val):
            raise RuntimeError("NLL is inf")
        return val

    def get_minuit_args(self):
        return (self.param_inits,)

    def get_minuit_kwargs(self):
        return dict(
            name=self.param_names,
            fix=self.param_fixes,
            limit=self.param_limit,
        )

    def create_minuit(self, **kwargs):
        _kwargs = copy.deepcopy(self.get_minuit_kwargs())
        _kwargs.update({"errordef": 1, "pedantic": False})
        _kwargs.update(kwargs)
        return iminuit.Minuit.from_array_func(
            self, *self.get_minuit_args(), **_kwargs,
        )

    def set_parameters(self, params, init=False):
        if isinstance(params, dict):
            params = np.array([
                params.get(name, init)
                for init, name in zip(self.param_inits, self.param_names)
            ])

        if np.isnan(params).any():
            raise RuntimeError(
                "NaN in input parameters. Previous values:\n{}\nNew values:\n{}".format(
                    self.parameters, params,
                )
            )
        if np.isinf(params).any():
            raise RuntimeError(
                "Inf in input parameters. Previous values:\n{}\nNew values:\n{}".format(
                    self.parameters, params,
                )
            )

        for name, val in zip(self.param_names, params):
            self.parameters[name] = val

        self.theta_mcstat = {}
        for r in self.regions:
            self.theta_mcstat[r] = []
            for name in self.mcstat_nuisances[r]:
                self.theta_mcstat[r].append(self.parameters[name])
            self.theta_mcstat[r] = np.array(self.theta_mcstat[r])

        self.theta_gaus = []
        self.theta_norm = []
        self.theta_morph = []
        for name in set(self.norm_nuisances+self.morph_nuisances):
            self.theta_gaus.append(self.parameters[name])
        for name in self.norm_nuisances:
            self.theta_norm.append(self.parameters[name])
        for name in self.morph_nuisances:
            self.theta_morph.append(self.parameters[name])
        self.theta_gaus = np.array(self.theta_gaus)
        self.theta_norm = np.array(self.theta_norm)
        self.theta_morph = np.array(self.theta_morph)

        if init:
            self.param_inits[:] = params

    def __call__(self, params, *args):
        if len(args)>0:
            params = np.array([params]+list(args))
        self.set_parameters(params)
        return self.nll()

    def fit(self, asimov=False, toy=-1, migrad=False, minos=False, **kwargs):
        self.asimov = asimov
        self.data_asimov = {
            r: np.maximum(0., self._predictions(r))
            for r in self.regions
        }

        np.random.seed(self.seed+toy)
        self.toy = toy
        self.data_toy = {
            r: scipy.stats.poisson.rvs(np.maximum(1e-10, self._predictions(r)))
            for r in self.regions
        }

        minimizer = self.create_minuit(**kwargs)
        if migrad:
            minimizer.migrad()
        if minos:
            minimizer.minos()
        self.minimizer = minimizer
        return minimizer

    def draw_data(self, region, ax, ratio=False, pull=False, **kwargs):
        data = self.data[region].copy()
        lower, upper = poisson_interval(data)
        if not pull:
            if ratio:
                mc = self.predictions(region)["prediction"].copy()
                data /= mc
                lower /= mc
                upper /= mc

            kw = dict(fmt='o', color='black')
            kw.update(kwargs)
            ax.errorbar(
                sum(self.bins[region])/2., data,
                yerr=(data-lower, upper-data), **kw,
            )
        else:
            mc = self.predictions(region)["prediction"].copy()
            mclower, mcupper = confidence_interval(self, self.minimizer, "_predictions", region)
            diff = data - mc
            sigma = np.sqrt((data-lower)**2+(mc-mcupper)**2)
            sigma[diff<0.] = np.sqrt((data-upper)**2+(mc-lower)**2)[diff<0.]

            kw = dict(color='black')
            kw.update(kwargs)
            ax.plot(sum(self.bins[region])/2., diff/sigma, 'o', **kw)

    def draw_prediction_total(
        self, region, ax, ratio=False, prefit=False, band=True, hkwargs={},
        fkwargs={},
    ):
        if prefit:
            saved_params = copy.deepcopy(self.parameters)
            self.set_parameters(self.param_inits)
            mc_prefit = self.predictions(region)["prediction"].copy()
            self.set_parameters(saved_params)
        mc = self.predictions(region)["prediction"].copy()
        lower, upper = confidence_interval(self, self.minimizer, "_predictions", region)
        if ratio:
            lower /= mc
            upper /= mc
            if prefit:
                mc_prefit /= mc
                mc = mc_prefit
            else:
                mc = np.ones_like(mc)

        kw = dict(histtype='step', color='black')
        if prefit:
            kw["ls"] = "--"
        kw.update(hkwargs)
        ax.hist(
            self.bins[region][0],
            bins=list(self.bins[region][0])+[self.bins[region][1][-1]],
            weights=mc, **kw,
        )

        if band:
            kw = dict(step='post', color='#d9d9d9')
            kw.update(fkwargs)
            ax.fill_between(
                list(self.bins[region][0])+[self.bins[region][1][-1]],
                list(lower)+[lower[-1]], list(upper)+[upper[-1]], **kw,
            )

    def draw_prediction_procs(self, region, ax, **kwargs):
        mc_procs = self.prediction(region).copy()
        procs = self.region_procs[region][::-1]

        kw = dict(
            stacked=True, color=[process_colours.get(p, 'blue') for p in procs],
            label=[process_names.get(p, p) for p in procs],
        )
        kw.update(kwargs)

        #w = mc_procs.reorder_levels(["bin_min", "bin_max", "process"])
        ws = [mc_procs.loc[(p,)] for p in procs]
        ax.hist(
            [self.bins[region][0]]*len(procs),
            bins=list(self.bins[region][0])+[self.bins[region][1][-1]],
            weights=ws, **kw,
        )

class NLLModel2D(NLLModel):
    def __init__(
        self, data, mc, config, same_bin_widths=False,
        shape_type="shape", smooth_region=1.,
        observable_function = "bin_min, bin_max: (bin_min + bin_max)/2."
    ):
        self.verbose = False
        self.dfdata = data
        self.dfmc = mc
        self.config = config

        self.shape_type = shape_type
        self.smooth_region = smooth_region

        self.asimov = False
        self.toy = -1
        self.seed = 123456
        self.saturated = False

        self.norm_nuisances = [
            p["name"]
            for p in config["parameters"]
            if p["constraint"] == "gaussian"
        ]
        self.morph_nuisances = [
            p["name"]
            for p in config["parameters"]
            if p["constraint"] == "gaussian"
        ]

        region_procs = config["regions"]
        self.region_procs = region_procs
        self.regions = list(region_procs.keys())

        self.mcstat_nuisances = {
            r: [p["name"] for p in config["parameters"] if r in p["name"] and p["constraint"]=="gamma"]
            for r in self.regions
        }

        self.param_names = []
        self.param_inits = []
        self.param_fixes = []
        self.param_limit = []
        for idx, params in enumerate(config["parameters"]):
            self.param_names.append(params["name"])
            self.param_inits.append(params["value"])
            self.param_fixes.append(params["fixed"])
            self.param_limit.append(params["limit"])

        self.bins = {}
        self.data = {}
        self.xobs= {}
        self.yobs= {}
        self.dbins = {}
        self.nbins = {}
        for r in self.regions:
            self.data[r] = data.loc[(r,),"sum_w"].values

            bins = data.loc[(r,),:].reset_index()[["bin0_min", "bin0_max", "bin1_min", "bin1_max"]].values
            self.xobs[r] = eval("lambda "+observable_function)(bins[:,2], bins[:,3])
            self.yobs[r] = eval("lambda "+observable_function)(bins[:,0], bins[:,1])

            self.dbins[r] = bins[:,3] - bins[:,2]
            if same_bin_widths:
                self.dbins[r][:] = 1.
            self.bins[r] = (bins[:,0], bins[:,1], bins[:,2], bins[:,3])

            nbins0 = data.loc[(r,),:].unstack().unstack().shape[0]
            nbins1 = data.loc[(r,),:].shape[0]//nbins0
            self.nbins[r] = (nbins0, nbins1)

        # self.data_asimov[r,b]
        # self.data_toy[r,b]

        self.scale_functions = {}
        for r in self.regions:
            self.scale_functions[r] = []
            functions = []
            for p in region_procs[r]:
                self.scale_functions[r].append(
                    config["scale_functions"].get((r, p), "x, y, w, p: np.ones_like(x)")
                )

        self.sumw = {}
        self.sumww = {}
        for r in self.regions:
            self.sumw[r] = []
            self.sumww[r] = []
            for p in region_procs[r]:
                self.sumw[r].append(mc.loc[(r,p,""),"sum_w"].values)
                self.sumww[r].append(mc.loc[(r,p,""),"sum_ww"].values)
            self.sumw[r] = np.array(self.sumw[r])
            self.sumww[r] = np.array(self.sumww[r])

        self.nom_sum = {}
        self.up_sum = {}
        self.do_sum = {}
        for r in self.regions:
            self.nom_sum[r] = []
            self.up_sum[r] = []
            self.do_sum[r] = []
            for p in region_procs[r]:
                self.nom_sum[r].append(mc.loc[(r,p,""),"sum_w"].sum())

                self.up_sum[r].append([])
                self.do_sum[r].append([])
                for n in self.norm_nuisances:
                    try:
                        self.up_sum[r][-1].append(
                            mc.loc[(r,p,n+"Up"),"sum_w"].sum()
                        )
                        self.do_sum[r][-1].append(
                            mc.loc[(r,p,n+"Down"),"sum_w"].sum()
                        )
                    except KeyError:
                        self.up_sum[r][-1].append(self.nom_sum[r][-1])
                        self.do_sum[r][-1].append(self.nom_sum[r][-1])
            self.nom_sum[r] = np.array(self.nom_sum[r])
            self.up_sum[r] = np.array(self.up_sum[r])
            self.do_sum[r] = np.array(self.do_sum[r])

        self.delta_sum = {}
        self.delta_diff = {}
        for r in self.regions:
            self.delta_sum[r] = []
            self.delta_diff[r] = []
            for p in region_procs[r]:
                self.delta_sum[r].append([])
                self.delta_diff[r].append([])

                nom = mc.loc[(r, p, ""), "sum_w"].values
                nom_norm = nom/(nom*self.dbins[r]).sum()
                for n in self.morph_nuisances:
                    skip = False
                    try:
                        up = mc.loc[(r,p,n+"Up"),"sum_w"].values
                        do = mc.loc[(r,p,n+"Down"),"sum_w"].values
                    except KeyError:
                        delta_sum = np.zeros_like(nom_norm)
                        delta_diff = np.zeros_like(nom_norm)
                        skip = True

                    if not skip:
                        up_norm = up/(up*self.dbins[r]).sum()
                        do_norm = do/(do*self.dbins[r]).sum()

                        if self.shape_type == "shape":
                            up_del = up_norm - nom_norm
                            do_del = do_norm - nom_norm
                        elif self.shape_type == "shapeN":
                            up_del = np.where(
                                (up_norm>0)&(nom_norm>0),
                                np.log(up_norm/nom_norm),
                                np.zeros_like(nom_norm),
                            )
                            do_del = np.where(
                                (do_norm>0)&(nom_norm>0),
                                np.log(do_norm/nom_norm),
                                np.zeros_like(nom_norm),
                            )
                        delta_diff = up_del - do_del
                        delta_sum = up_del + do_del
                    self.delta_diff[r][-1].append(delta_diff)
                    self.delta_sum[r][-1].append(delta_sum)
            self.delta_diff[r] = np.swapaxes(np.array(self.delta_diff[r]), 1, 2)
            self.delta_sum[r] = np.swapaxes(np.array(self.delta_sum[r]), 1, 2)

        self.sumw_procsum = {
            r: (
                mc.loc[(r, pd.IndexSlice[:], ""), "sum_w"]
                .groupby(["bin0_min", "bin0_max", "bin1_min", "bin1_max"]).sum().values
            ) for r in self.regions
        }
        self.sumww_procsum = {
            r: (
                mc.loc[(r, pd.IndexSlice[:], ""), "sum_ww"]
                .groupby(["bin0_min", "bin0_max", "bin1_min", "bin1_max"]).sum().values
            ) for r in self.regions
        }
        self.neff_procsum = {
            r: self.sumw_procsum[r]**2/self.sumww_procsum[r]
            for r in self.regions
        }

        self.parameters = {}
        self.set_parameters({p["name"]: p["value"] for p in config["parameters"]})

    def _prediction(self, region):
        nom_scale = np.ones_like(self.sumw[region])
        for idx, function in enumerate(self.scale_functions[region]):
            nbins = self.nbins[region]
            xobs = self.xobs[region].reshape(nbins)
            yobs = self.yobs[region].reshape(nbins)
            sumw = self.sumw[region][idx,:].reshape(nbins)
            nom_scale[idx,:] = eval('lambda '+function)(
                xobs, yobs, sumw, self.parameters,
            ).ravel()

        syst_scale = self._shape_morphs(region)*(self._shape_norms(region)[:,np.newaxis])

        #mcstat_scale = (1 + np.sqrt(self.sumww)/self.sumw)**(self.theta_mcstat[:,np.newaxis,:])
        mcstat_scale = np.power(
            (1 + np.sqrt(self.sumww_procsum[region])/self.sumw_procsum[region]),
            self.theta_mcstat[region]
        )[np.newaxis,:]

        # return sumw*nom_scale*syst_scale*mcstat_scale
        # syst_scale already includes sumw
        return nom_scale*syst_scale*mcstat_scale

    def draw_data(self, region, ax, ratio=False, pull=False, **kwargs):
        data = self.data[region].copy()
        lower, upper = poisson_interval(data)
        if not pull:
            if ratio:
                mc = self.predictions(region)["prediction"].copy()
                data /= mc
                lower /= mc
                upper /= mc

            kw = dict(fmt='o', color='black')
            kw.update(kwargs)
            ax.errorbar(
                sum(self.bins[region])/2., data,
                yerr=(data-lower, upper-data), **kw,
            )
        else:
            mc = self.predictions(region)["prediction"].copy()
            mclower, mcupper = confidence_interval(self, self.minimizer, "_predictions", region)
            diff = data - mc
            sigma = np.sqrt((data-lower)**2+(mc-mcupper)**2)
            sigma[diff<0.] = np.sqrt((data-upper)**2+(mc-lower)**2)[diff<0.]

            kw = dict(color='black')
            kw.update(kwargs)
            ax.plot(sum(self.bins[region])/2., diff/sigma, 'o', **kw)

def run_minos(
    params, nll_args=tuple(), nll_kwargs={}, migrad_kwargs={"ncall": 100000},
    fit_kwargs={"migrad": False, "minos": False},
    pysge_function="local_submit", pysge_args=tuple(), pysge_kwargs={},
    params_guess=None, nll_class=NLLModel, fixed_params=[],
):
    def run_minos_param(
        _param, _nll_args, _nll_kwargs, _migrad_kwargs,
        _param_guess=None,
    ):
        mod_= nll_class(*_nll_args, **_nll_kwargs)
        mini_ = mod_.fit(**fit_kwargs)
        if _param_guess is not None:
            for k, v in _param_guess.items():
                mini_.values[k] = v
        if len(fixed_params)>0:
            for k in fixed_params:
                mini_.fixed[k] = True
        mini_.migrad(**_migrad_kwargs)
        mini_.migrad(**_migrad_kwargs)
        valid_nom = mini_.migrad_ok()
        if not valid_nom:
            mini_.migrad(**_migrad_kwargs)
            valid_nom = mini_.migrad_ok()
        mini_.minos(_param)
        return {
            "parameter": _param,
            "value": mini_.values[_param],
            "merror_up": mini_.merrors[(_param, 1.)],
            "merror_down": mini_.merrors[(_param, -1.)],
            "upper_valid": mini_.get_merrors()[_param].upper_valid,
            "lower_valid": mini_.get_merrors()[_param].lower_valid,
        }

    tasks = [{
        "task": run_minos_param,
        "args": (p, nll_args, nll_kwargs, migrad_kwargs),
        "kwargs": {"_param_guess": params_guess},
    } for p in params]

    results = getattr(pysge, pysge_function)(
        *tuple([tasks]+list(pysge_args)),
        **pysge_kwargs,
    )
    return pd.DataFrame(results)

def run_impacts(
    params, poi, nll_args=tuple(), nll_kwargs={},
    fit_kwargs={"migrad": False, "minos": False},
    migrad_kwargs={"ncall": 100000}, pysge_function="local_submit",
    pysge_args=tuple(), pysge_kwargs={}, params_guess=None,
    return_tasks=False, nll_class=NLLModel,
):
    def run_impact_param(
        _param, _poi, _nll_args, _nll_kwargs, _fit_kwargs, _migrad_kwargs,
        _param_guess=None,
    ):
        mod_= nll_class(*_nll_args, **_nll_kwargs)
        mini_ = mod_.fit(**_fit_kwargs)
        if _param_guess is not None:
            for k, v in _param_guess.items():
                mini_.values[k] = v
        mini_.migrad(**_migrad_kwargs)
        valid_nom = mini_.migrad_ok()
        if not valid_nom:
            mini_.migrad(**_migrad_kwargs)
            valid_nom = mini_.migrad_ok()
        mini_.minos(_param)

        if _param == _poi:
            return {
                "param": _param,
                "param_value": mini_.values[_param],
                "param_merrup": mini_.merrors[(_param, 1.)],
                "param_merrdown": mini_.merrors[(_param, -1.)],
                "poi": _poi,
                "poi_paramup": mini_.merrors[(_param, 1.)],
                "poi_paramdown": mini_.merrors[(_param, -1.)],
            }

        poi_bf_central = copy.deepcopy(mini_.values[_poi])
        param_bf_central = copy.deepcopy(mini_.values[_param])
        param_bf_up = copy.deepcopy(mini_.merrors[(_param, 1.)])
        param_bf_down = copy.deepcopy(mini_.merrors[(_param, -1.)])

        # Perform fit at +1 sigma
        mini_.values[_param] = param_bf_central + param_bf_up
        mini_.fixed[_param] = True
        mini_.migrad(**_migrad_kwargs)
        valid_up = mini_.migrad_ok()
        if not valid_up:
            mini_.migrad(**_migrad_kwargs)
            valid_up = mini_.migrad_ok()
        poi_bf_nuisance_up = copy.deepcopy(mini_.values[_poi])

        # Perform fit at -1 sigma
        mini_.values[_param] = param_bf_central + param_bf_down
        mini_.migrad(**_migrad_kwargs)
        valid_do = mini_.migrad_ok()
        if not valid_do:
            mini_.migrad(**_migrad_kwargs)
            valid_do = mini_.migrad_ok()
        poi_bf_nuisance_down = copy.deepcopy(mini_.values[_poi])

        return {
            "param": _param,
            "param_value": param_bf_central,
            "param_merrup": param_bf_up,
            "param_merrdown": param_bf_down,
            "poi": _poi,
            "poi_paramup": poi_bf_nuisance_up-poi_bf_central,
            "poi_paramdown": poi_bf_nuisance_down-poi_bf_central,
            "valid": valid_nom and valid_up and valid_do,
        }

    tasks = [{
        "task": run_impact_param,
        "args": (p, poi, nll_args, nll_kwargs, fit_kwargs, migrad_kwargs),
        "kwargs": {"_param_guess": params_guess},
    } for p in params]
    if return_tasks:
        return tasks

    results = getattr(pysge, pysge_function)(
        *tuple([tasks]+list(pysge_args)),
        **pysge_kwargs,
    )
    return pd.DataFrame(results)

def run_profile_likelihood(
    param, param_values, nll_args=tuple(), nll_kwargs={},
    fit_kwargs={"migrad": False, "minos": False},
    migrad_kwargs={"ncall": 100000}, pysge_function="local_submit",
    pysge_args=tuple(), pysge_kwargs={}, params_guess=None, nll_class=NLLModel,
    fixed_params=[],
):
    def run_fit(
        _param, _paramval, _nll_args, _nll_kwargs, _migrad_kwargs,
        _param_guess=None,
    ):
        mod_ = nll_class(*_nll_args, **_nll_kwargs)
        mini_ = mod_.fit(**fit_kwargs)
        if _param_guess is not None:
            for k, v in _param_guess.items():
                mini_.values[k] = v
        if len(fixed_params):
            for k in fixed_params:
                mini_.fixed[k] = True
        mini_.migrad(**_migrad_kwargs)
        valid = mini_.migrad_ok()
        if not valid:
            mini_.migrad(**_migrad_kwargs)
            valid = mini_.migrad_ok()
        param_bf = copy.deepcopy(mini_.values[_param])
        nll_bf = copy.deepcopy(mini_.get_fmin().fval)

        mini_.values[_param] = _paramval
        mini_.fixed[_param] = True
        mini_.migrad(**_migrad_kwargs)
        valid = mini_.migrad_ok()
        if not valid:
            mini_.migrad(**_migrad_kwargs)
            valid = mini_.migrad_ok()
        nll_alt = copy.deepcopy(mini_.get_fmin().fval)

        return {
            "parameter": _param,
            "parameter_bf": param_bf,
            "parameter_value": _paramval,
            "dnll": nll_alt - nll_bf,
        }

    tasks = [{
        "task": run_fit,
        "args": (param, pv, nll_args, nll_kwargs, migrad_kwargs),
        "kwargs": {"_param_guess": params_guess},
    } for pv in param_values]

    results = getattr(pysge, pysge_function)(
        *tuple([tasks]+list(pysge_args)),
        **pysge_kwargs,
    )
    return pd.DataFrame(results)

def run_profile_likelihood_nd(
    params, param_values, nll_args=tuple(), nll_kwargs={},
    fit_kwargs={"migrad": False, "minos": False},
    migrad_kwargs={"ncall": 100000}, pysge_function="local_submit",
    pysge_args=tuple(), pysge_kwargs={}, params_guess=None, nll_class=NLLModel,
    fixed_params=[],
):
    def run_fit(
        _params, _paramsval, _nll_args, _nll_kwargs, _migrad_kwargs,
        _param_guess=None,
    ):
        mod_ = nll_class(*_nll_args, **_nll_kwargs)
        mini_ = mod_.fit(**fit_kwargs)
        if _param_guess is not None:
            for k, v in _param_guess.items():
                mini_.values[k] = v
        if len(fixed_params):
            for k in fixed_params:
                mini_.fixed[k] = True
        mini_.migrad(**_migrad_kwargs)
        valid = mini_.migrad_ok()
        if not valid:
            mini_.migrad(**_migrad_kwargs)
            valid = mini_.migrad_ok()
        params_bf = {
            p: copy.deepcopy(mini_.values[p])
            for p in _params
        }
        nll_bf = copy.deepcopy(mini_.get_fmin().fval)

        for p, pv in zip(_params, _paramsval):
            mini_.values[p] = pv
            mini_.fixed[p] = True
        mini_.migrad(**_migrad_kwargs)
        valid = mini_.migrad_ok()
        if not valid:
            mini_.migrad(**_migrad_kwargs)
            valid = mini_.migrad_ok()
        nll_alt = copy.deepcopy(mini_.get_fmin().fval)

        results = {}
        for idx, p in enumerate(_params):
            results["p{}".format(idx)] = p
            results["pbf{}".format(idx)] = params_bf[p]
            results["pv{}".format(idx)] = _paramsval[idx]
        results["dnll"] = nll_alt - nll_bf
        return results

    tasks = []
    for pv0 in param_values[0]:
        for pv1 in param_values[1]:
            tasks.append({
                "task": run_fit,
                "args": (params, (pv0, pv1), nll_args, nll_kwargs, migrad_kwargs),
                "kwargs": {"_param_guess": params_guess},
            })

    results = getattr(pysge, pysge_function)(
        *tuple([tasks]+list(pysge_args)),
        **pysge_kwargs,
    )
    return pd.DataFrame(results)

def confidence_interval(model, minimizer, attr, *args, **kwargs):
    def pred_func(params):
        old_params = copy.deepcopy(model.parameters)
        model.set_parameters(params)
        result = getattr(model, attr)(*args, **kwargs)
        model.set_parameters(old_params)
        return result

    bf_params = minimizer.np_values()

    cov = minimizer.np_matrix(skip_fixed=False)
    jac = numdifftools.Jacobian(pred_func)(bf_params)
    err = scipy.stats.chi2.ppf(0.68, df=1)*np.sqrt(
        np.matmul(jac, np.matmul(cov, jac.T)).diagonal()
    )
    cent = pred_func(bf_params)
    return cent-err, cent+err



import numpy as np

def _rect_inter_inner(x1,x2):
    n1=x1.shape[0]-1
    n2=x2.shape[0]-1
    X1=np.c_[x1[:-1],x1[1:]]
    X2=np.c_[x2[:-1],x2[1:]]
    S1=np.tile(X1.min(axis=1),(n2,1)).T
    S2=np.tile(X2.max(axis=1),(n1,1))
    S3=np.tile(X1.max(axis=1),(n2,1)).T
    S4=np.tile(X2.min(axis=1),(n1,1))
    return S1,S2,S3,S4

def _rectangle_intersection_(x1,y1,x2,y2):
    S1,S2,S3,S4=_rect_inter_inner(x1,x2)
    S5,S6,S7,S8=_rect_inter_inner(y1,y2)

    C1=np.less_equal(S1,S2)
    C2=np.greater_equal(S3,S4)
    C3=np.less_equal(S5,S6)
    C4=np.greater_equal(S7,S8)

    ii,jj=np.nonzero(C1 & C2 & C3 & C4)
    return ii,jj

def intersection(x1,y1,x2,y2):
    ii,jj=_rectangle_intersection_(x1,y1,x2,y2)
    n=len(ii)

    dxy1=np.diff(np.c_[x1,y1],axis=0)
    dxy2=np.diff(np.c_[x2,y2],axis=0)

    T=np.zeros((4,n))
    AA=np.zeros((4,4,n))
    AA[0:2,2,:]=-1
    AA[2:4,3,:]=-1
    AA[0::2,0,:]=dxy1[ii,:].T
    AA[1::2,1,:]=dxy2[jj,:].T

    BB=np.zeros((4,n))
    BB[0,:]=-x1[ii].ravel()
    BB[1,:]=-x2[jj].ravel()
    BB[2,:]=-y1[ii].ravel()
    BB[3,:]=-y2[jj].ravel()

    for i in range(n):
        try:
            T[:,i]=np.linalg.solve(AA[:,:,i],BB[:,i])
        except:
            T[:,i]=np.NaN


    in_range= (T[0,:] >=0) & (T[1,:] >=0) & (T[0,:] <=1) & (T[1,:] <=1)

    xy0=T[2:,in_range]
    xy0=xy0.T
    return xy0[:,0],xy0[:,1]




import numpy as np
import scipy.stats

__all__ = [
    "fc_poisson_efficiency_interval",
    "poisson_interval",
]

def fc_poisson_efficiency_interval(total, passed, level=0.68):
    import ROOT
    ROOT.gROOT.SetBatch(True)

    up = np.zeros(total.shape, dtype='float')
    down = np.zeros(total.shape, dtype='float')
    for idx, (tot, pas) in enumerate(zip(total, passed)):
        up[idx] = ROOT.TEfficiency.FeldmanCousins(tot, pas, level, True)
        down[idx] = ROOT.TEfficiency.FeldmanCousins(tot, pas, level, False)
    return down, up

def poisson_interval(counts, scale=1, level=0.68):
    """
    Determine the up and down variations for the poisson interval (possibly
    scaled).

    Parameters
    ----------
    counts : ndarray-like
        The counts or effective number of entries in histogrammed bins

    scale : ndarray-like
        Scale to give to the gamma distribution interval for weighted events.
        Should be equivalent to sum_w/neff.

    level : float (default 0.68)
        The coverage interval. Default is 68%, i.e. near 1 sigma.
    """
    if np.isscalar(counts):
        _counts = np.array([counts])
    elif isinstance(counts, list):
        _counts = np.array(counts)
    else:
        _counts = counts

    l = scipy.stats.gamma.interval(level, _counts, scale=scale)[0]
    l[_counts==0] = 0.
    u = scipy.stats.gamma.interval(level, _counts+1, scale=scale)[1]

    ret = (l, u)
    if np.isscalar(counts):
        ret = (l[0], u[0])
    return ret



from scipy.ndimage import gaussian_filter1d
from scipy.interpolate import UnivariateSpline

def unispline_to_gausfilter(
    x, y, w=None, spline_kw={"k": 2}, filter_kw={"sigma": 1},
):
    if w is not None:
        spline_kw["w"] = w

    s = UnivariateSpline(x, y, **spline_kw)
    return gaussian_filter1d(s(x), **filter_kw)



import setuptools

with open("README.md", 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name="dftools",
    version="0.2.3",
    author="Shane Breeze",
    author_email="sdb15@ic.ac.uk",
    scripts=[],
    description="dataframe tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        "numpy>=1.18.1",
        "pandas>=1.0.1",
        "matplotlib>=3.1.2",
        "scipy>=1.4.1",
        "numdifftools>=0.9.39",
        "iminuit>=1.3.8",
    ],
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Development Status :: 3 - Alpha",
    ],
)



import functools
import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.transforms as transforms
import matplotlib.pyplot as plt
import scipy.interpolate as interp
import scipy.optimize as opt
from .stats import poisson_interval

__all__ = [
    "cms_label", "legend_data_mc", "data_mc", "data", "mc", "heatmap",
    "annotate_heatmap",
    "process_names", "process_colours",
    "impacts", "nllscan",
]

def cms_label(ax, label, lumi=35.9, energy=13):
    ax.text(
        0, 1, r'$\mathbf{CMS}\ \mathit{'+label+'}$',
        ha='left', va='bottom', transform=ax.transAxes,
    )
    ax.text(
        1, 1, r'${:.1f}\ \mathrm{{fb}}^{{-1}}$ ({:.0f} TeV)'.format(lumi, energy),
        ha='right', va='bottom', transform=ax.transAxes,
    )

def legend_data_mc(
    ax, df_data, df_mc, label, add_ratios=True, offaxis=True, legend_kw={},
):
    handles, labels = ax[0].get_legend_handles_labels()

    if add_ratios:
        # sort by process total
        tdf_mc = pd.pivot_table(
            df_mc, index=label, columns="parent",
            values="sum_w", aggfunc=np.sum,
        )
        tdf_mc = tdf_mc[tdf_mc.sum(axis=0).sort_values().index]

        data_idx = labels.index("Data")
        data_label = labels.pop(data_idx)
        labels = (labels+[data_label])[::-1]
        data_handle = handles.pop(data_idx)
        handles = (handles+[data_handle])[::-1]

        df_data_sum = df_data.sum()
        tdf_mc_sum = tdf_mc.sum()

        fractions = [
            df_data_sum["sum_w"]/tdf_mc_sum.sum(), 1.,
        ] + list((tdf_mc_sum / tdf_mc_sum.sum()).values[::-1])
        fraction_labels = [
            "{:.3f} {}".format(fractions[idx], labels[idx])
            for idx in range(len(labels))
        ]
    else:
        handles = handles[::-1]
        fraction_labels = labels[::-1]

    kwargs = dict(legend_kw)
    kwargs_noloc = dict(kwargs)
    kwargs_noloc.pop("loc", None)
    if offaxis:
        box = ax[0].get_position()
        ax[0].set_position([box.x0, box.y0, box.width*0.8, box.height])
        ax[0].legend(
            handles, fraction_labels, bbox_to_anchor=(1, 1), **kwargs_noloc
        )
        box = ax[1].get_position()
        ax[1].set_position([box.x0, box.y0, box.width*0.8, box.height])
    else:
        ax[0].legend(handles, fraction_labels, **kwargs)

    handles, labels = ax[1].get_legend_handles_labels()
    if offaxis:
        ax[1].legend(handles, labels, bbox_to_anchor=(1, 1), **kwargs_noloc)
    else:
        ax[1].legend(handles, labels, **kwargs)

def bin_lows_to_edges_cents(lows):
    edges = np.array(list(lows)+[2*lows[-1]-lows[-2]])
    cents = (edges[:-1] + edges[1:])/2.
    return edges, cents

def data(ax, df, label, bins, data_kw={}):
    bin_edges, bin_cents = bin_lows_to_edges_cents(bins)

    # draw
    kwargs = dict(fmt='o', lw=1, color='black', label='Data')
    kwargs.update(data_kw)

    #down, up = poisson_interval(df["count"], scale=df["sum_w"]/df["count"])
    neff = df["sum_w"]**2 / df["sum_ww"]
    down, up = poisson_interval(neff, scale=df["sum_w"]/neff)
    ax.errorbar(
        bin_cents, df["sum_w"], yerr=[df["sum_w"]-down, up-df["sum_w"]],
        **kwargs,
    )

def mc(
    ax, df, label, bins, 
    mcstat=False, mc_kw={}, mcstat_kw={}, proc_kw={}, zorder=0,
):
    bin_edges, bin_cents = bin_lows_to_edges_cents(bins)

    # preprocess mc
    tdf = pd.pivot_table(
        df, index=label, columns="parent",
        values="sum_w", aggfunc=np.sum,
    )

    # sort by process total
    tdf_procsum = tdf.sum(axis=0)
    tdf = tdf[tdf_procsum.sort_values().index]

    # mc
    procs = tdf.columns.to_series()
    kwargs = dict(
        label=procs.replace(proc_kw.get("labels", {})),
        color=procs.apply(lambda x: proc_kw.get("colours", {}).get(x, "blue")),
        zorder=zorder,
    )
    kwargs.update(mc_kw)
    ax.hist([bin_cents]*tdf.shape[1], bins=bin_edges, weights=tdf.T, **kwargs)

    if mcstat:
        tdf_ww = pd.pivot_table(
            df, index=label, columns="parent",
            values="sum_ww", aggfunc=np.sum,
        )
        neff = tdf**2 / tdf_ww
        down, up = poisson_interval(neff, scale=tdf/neff)
        kwargs = dict(color='black', alpha=0.2)
        kwargs.update(mcstat_kw)
        up = up[:,0]
        down = down[:,0]
        ax.fill_between(
            bin_edges, list(up)+[up[-1]], list(down)+[down[-1]],
            step='post', **kwargs
        )

def data_mc(
    ax, df_data, df_mc, label, bins, sigs=[], blind=False, log=True, legend=True,
    ratio=True, sm_total=True, mcstat_top=False, add_ratios=True, mc_kw={},
    sig_kw={}, mcstat_kw={}, sm_kw={}, data_kw={}, proc_kw={}, legend_kw={}, 
    cms_kw={},
):
    if blind:
        df_data = df_data*0.

    # collect signals if set
    sigs = sigs[::-1]
    sig_mask = ~df_mc.index.get_level_values("parent").isin(sigs)
    df_sig = df_mc.loc[~sig_mask].copy(deep=True)

    df_mc_sm = df_mc.loc[sig_mask].copy(deep=True)

    # preprocessing
    df_mc_sum = df_mc_sm.groupby(label).sum()
    df_mc_sum.loc[:,"parent"] = "SMTotal"
    df_mc_sum = df_mc_sum.groupby(["parent", label]).sum()

    # draw
    if log:
        ax[0].set_yscale('log')

    bin_edges, _ = bin_lows_to_edges_cents(bins)
    ax[0].set_xlim(bin_edges.min(), bin_edges.max())

    # signals - top panel
    sig_kw_ = dict(histtype='step', zorder=1)
    sig_kw_.update(sig_kw)
    if len(sigs) > 0:
        mc(
            ax[0], df_sig, label, bins, mcstat=False, mc_kw=sig_kw_,
            proc_kw=proc_kw,
        )

    # MC - top panel
    mc_kw_ = dict(stacked=True)
    mc_kw_.update(mc_kw)
    mc(
        ax[0], df_mc_sm, label, bins, mcstat=mcstat_top, mc_kw=mc_kw_,
        proc_kw=proc_kw,
    )

    # SM total - top panel
    if sm_total:
        mc_kw_ = dict(histtype='step')
        mc_kw_.update(sm_kw)
        mcstat_kw_ = dict(label="", color="black", alpha=0.2)
        mcstat_kw_.update(mcstat_kw)
        mc(
            ax[0], df_mc_sum, label, bins, mcstat=False, mc_kw=mc_kw_,
            mcstat_kw=mcstat_kw_, proc_kw=proc_kw,
        )

    # Data - top panel
    data(ax[0], df_data, label, bins, data_kw=data_kw)

    # CMS label - top panel
    kwargs = dict(label="Preliminary", lumi=35.9, energy=13)
    kwargs.update(cms_kw)
    #cms_label(ax[0], **kwargs)

    # SM total ratio - bottom panel
    df_mc_sum_ratio = df_mc_sum.copy()
    #df_mc_sum_ratio.loc[:,"count"] = df_mc_sum["count"]
    df_mc_sum_ratio.loc[:,"sum_w"] = 1.
    df_mc_sum_ratio.loc[:,"sum_ww"] = df_mc_sum["sum_ww"]/df_mc_sum["sum_w"]**2

    if ratio:
        mc_kw_ = dict(label="", histtype='step')
        mc_kw_.update(sm_kw)
        mcstat_kw_ = dict(label="MC stat. unc.", color="black", alpha=0.2)
        mcstat_kw_.update(mcstat_kw)
        mc(
            ax[1], df_mc_sum_ratio, label, bins, mcstat=True, mc_kw=mc_kw_,
            mcstat_kw=mcstat_kw_, proc_kw=proc_kw,
        )

        # Data ratio - bottom panel
        if not blind:
            kwargs = dict(data_kw)
            kwargs["label"] = ""
            df_data_ratio = df_data.copy()
            #df_data_ratio.loc[:,"count"] = df_data["count"]
            df_data_ratio.loc[:,"sum_w"] = df_data["sum_w"]/df_mc_sum["sum_w"].values
            df_data_ratio.loc[:,"sum_ww"] = df_data["sum_ww"]/df_mc_sum["sum_w"].values**2
            data(ax[1], df_data_ratio, label, bins, data_kw=kwargs)

        if legend:
            offaxis = legend_kw.pop("offaxis", True)
            kwargs = dict(labelspacing=0.05)
            kwargs.update(legend_kw)
            legend_data_mc(
                ax, df_data, df_mc, label, add_ratios=add_ratios,
                offaxis=offaxis, legend_kw=kwargs,
            )

    return ax

def heatmap(
    data, row_labels, col_labels, ax, cbar_kw=dict(fraction=0.046, pad=0.04),
    cbarlabel="", grid_kw={}, tick_kw={}, **kwargs,
):
    if not ax:
        ax = plt.gca()

    im = ax.imshow(data, **kwargs)

    cbar = ax.figure.colorbar(im, ax=ax, **cbar_kw)
    cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")

    ax.set_xticks(np.arange(data.shape[1]))
    ax.set_yticks(np.arange(data.shape[0]))
    ax.set_xticklabels(col_labels)
    ax.set_yticklabels(row_labels)
    ax.tick_params(**tick_kw)

    # Rotate the tick labels and set their alignment.
    plt.setp(
        ax.get_xticklabels(), ha="right", #rotation=-30,
        rotation_mode="anchor",
    )

    # Turn spines off and create white grid.
    for edge, spine in ax.spines.items():
        spine.set_visible(False)

    ax.set_xticks(np.arange(data.shape[1]+1)-.5, minor=True)
    ax.set_yticks(np.arange(data.shape[0]+1)-.5, minor=True)

    gkw = dict(which="minor", color="w", linestyle='-', linewidth=2)
    gkw.update(grid_kw)
    ax.grid(**gkw)
    ax.tick_params(
        which="minor", bottom=False, left=False, top=False, right=False,
    )
    ax.tick_params(
        which="major", bottom=False, left=False, top=False, right=False,
    )

    return im, cbar

def annotate_heatmap(
    im, data=None, valfmt="{x:.2f}", textcolors=["black", "white"],
    cthreshold=lambda z: True, vthreshold=lambda z: True, **textkw,
):
    if not isinstance(data, (list, np.ndarray)):
        data = im.get_array()

    kw = dict(ha="center", va="center")
    kw.update(textkw)

    # Get the formatter in case a string is supplied
    if isinstance(valfmt, str):
        valfmt = mpl.ticker.StrMethodFormatter(valfmt)

    # Loop over the data and create a `Text` for each "pixel".
    # Change the text's color depending on the data.
    texts = []
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            kw.update(color=textcolors[int(cthreshold(data[i, j]))])
            if not vthreshold(data[i, j]):
                continue
            text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
            texts.append(text)

    return texts


process_colours = {
    "SMTotal":          "black",
    "MET":              "black",
    "SingleMuon":       "black",
    "SingleElectron":   "black",
    "ZJetsToNuNu":      "#80b1d3",
    "WJetsToLNu":       "#b3de69",
    "WJetsToENu":       "#b2df8a",
    "WJetsToMuNu":      "#b3de69",
    "WJetsToTauNu":     "#8dd3c7",
    "WJetsToTauLNu":    "#8dd3c7",
    "WJetsToTauHNu":    "#8dd3c7",
    "Diboson":          "#fdb462",
    "DYJetsToLL":       "#ffed6f",
    "DYJetsToEE":       "#fff6b3",
    "DYJetsToMuMu":     "#ffed6f",
    "DYJetsToTauTau":   "#ffe41a",
    "DYJetsToTauLTauL": "#ffe41a",
    "DYJetsToTauHTauL": "#ffe41a",
    "DYJetsToTauHTauH": "#ffe41a",
    "EWKV2Jets":        "#bebada",
    "SingleTop":        "#fccde5",
    "TTJets":           "#bc80bd",
    "Top":              "#bc80bd",
    "QCD":              "#fb8072",
    "G1Jet":            "#ccebc5",
    "VGamma":           "#ffffb3",
    "Minor":            "#d9d9d9",
    "MinorBkgs":        "#d9d9d9",
}

process_names = {
    "SMTotal":          "SM total",
    "MET":              "MET",
    "SingleMuon":       "Single Muon",
    "SingleElectron":   "Single Electron",
    "ZJetsToNuNu":      "$Z(\\rightarrow \\nu\\nu)+j$",
    "WJetsToLNu":       "$W(\\rightarrow l\\nu)+j$",
    "WJetsToENu":       "$W(\\rightarrow e\\nu)+j$",
    "WJetsToMuNu":      "$W(\\rightarrow \\mu\\nu)+j$",
    "WJetsToTauNu":     "$W(\\rightarrow \\tau\\nu)+j$",
    "WJetsToTauLNu":    "$W(\\rightarrow \\tau_{l}\\nu)+j$",
    "WJetsToTauHNu":    "$W(\\rightarrow \\tau_{h}\\nu)+j$",
    "Diboson":          "Diboson",
    "DYJetsToLL":       "$Z/\\gamma^{*}(\\rightarrow ll)+j$",
    "DYJetsToEE":       "$Z/\\gamma^{*}(\\rightarrow ee)+j$",
    "DYJetsToMuMu":     "$Z/\\gamma^{*}(\\rightarrow \\mu\\mu)+j$",
    "DYJetsToTauTau":   "$Z/\\gamma^{*}(\\rightarrow \\tau\\tau)+j$",
    "DYJetsToTauLTauL": "$Z/\\gamma^{*}(\\rightarrow \\tau_{l}\\tau_{l})+j$",
    "DYJetsToTauHTauL": "$Z/\\gamma^{*}(\\rightarrow \\tau_{l}\\tau_{h})+j$",
    "DYJetsToTauHTauH": "$Z/\\gamma^{*}(\\rightarrow \\tau_{h}\\tau_{h})+j$",
    "EWKV2Jets":        "VBS",
    "SingleTop":        "Single Top",
    "TTJets":           "$t\\bar{t}+j$",
    "QCD":              "QCD multijet",
    "G1Jet":            "$\\gamma+j$",
    "VGamma":           "$V+\\gamma$",
    "Minor":            "Minor",
    "MinorBkgs":        "Minor",
}

nuisance_names = {
    "d1kqcd": r'$\delta^{(1)}k_{\mathrm{QCD}}$',
    "d2kqcd": r'$\delta^{(2)}k_{\mathrm{QCD}}$',
    "d3kqcd": r'$\delta^{(3)}k_{\mathrm{QCD}}$',
    "d1kew": r'$\delta^{(1)}k_{\mathrm{EW}}$',
    "d2keww": r'$\delta^{(2)}k_{\mathrm{EW}}^{\mathrm{W}}$',
    "d2kewz": r'$\delta^{(2)}k_{\mathrm{EW}}^{\mathrm{Z}}$',
    "d3keww": r'$\delta^{(3)}k_{\mathrm{EW}}^{\mathrm{W}}$',
    "d3kewz": r'$\delta^{(3)}k_{\mathrm{EW}}^{\mathrm{Z}}$',
    "dkmix": r'$\delta k_{\mathrm{mix}}$',
    "jesTotal": r'JES',
    "jerSF": r'JER',
    "unclust": r'Unclustered energy',
    "lhePdfWeight": r'PDF',
    "btagSF": r'$b$-tag veto',
    "photonIdLoose": r'Photon id. veto',
    "photonPixelSeedVeto": r'Photon pixel veto',
    "tauIdTight": r'$\tau_h$-tag id. selection',
    "tauIdVLoose": r'$\tau_h$-tag id. veto',
    "muonIdLooseSyst": r'Muon id. veto (syst.)',
    "muonIdLooseStat": r'Muon id. veto (stat.)',
    "muonIsoLooseSyst": r'Muon iso. veto (syst.)',
    "muonIsoLooseStat": r'Muon iso. veto (stat.)',
    "muonIdTightSyst": r'Muon id. selection (syst.)',
    "muonIdTightStat": r'Muon id. selection (stat.)',
    "muonIsoTightSyst": r'Muon iso. selection (syst.)',
    "muonIsoTightStat": r'Muon iso. selection (stat.)',
    "eleIdIsoVeto": r'Electron id. veto',
    "eleIdIsoTight": r'Electron id. selection',
    "eleReco": r'Electron reconstruction',
    "eleTrig": r'Electron trigger',
    "prefiring": r'ECAL timing',
    "pileup": r'Pileup',
    "lumi": r'Luminosity',
    "metTrig0MuSyst": r'$p_{\mathrm{T}}^{\mathrm{miss}}$ trigger ($0\mu$)',
    "metTrig1MuSyst": r'$p_{\mathrm{T}}^{\mathrm{miss}}$ trigger ($1\mu$)',
    "metTrig2MuSyst": r'$p_{\mathrm{T}}^{\mathrm{miss}}$ trigger ($2\mu$)',
    "metTrigReferenceTriggerSyst": r'$p_{\mathrm{T}}^{\mathrm{miss}}$ trigger (ref.)',
    "metTrigMonojetSyst": r'$p_{\mathrm{T}}^{\mathrm{miss}}$ trigger ($\p_{\mathrm{T}}^{\mathrm{miss}}+\mathrm{jets}$)',
    "metTrigSingleMuonSyst": r'$p_{\mathrm{T}}^{\mathrm{miss}}$ trigger ($\mu+\mathrm{jets}$)',
    "metTrigDoubleMuonSyst": r'$p_{\mathrm{T}}^{\mathrm{miss}}$ trigger ($\mu\mu+\mathrm{jets}$)',
    "metTrigSingleTauSyst": r'$p_{\mathrm{T}}^{\mathrm{miss}}$ trigger ($\tau_h+\mathrm{jets}$)',
    "metTrigSingleElectronSyst": r'$p_{\mathrm{T}}^{\mathrm{miss}}$ trigger ($e+\mathrm{jets}$)',
    "metTrigDoubleElectronSyst": r'$p_{\mathrm{T}}^{\mathrm{miss}}$ trigger ($ee+\mathrm{jets}$)',
}

def impacts(data, ax=None, converter=nuisance_names):
    if ax is None:
        fig, ax = plt.subplots(
            figsize=(4,4), dpi=150,
            ncols=2, nrows=1,
            sharex=False, sharey=True,
            gridspec_kw={"hspace": 0., "wspace": 0.},
        )

    ax[0].minorticks_off()
    ax[1].minorticks_off()
    ax[1].set_yticklabels([])

    y = data["poi_paramdown"].values
    x = np.linspace(0., len(y), len(y)+1)
    ax[1].hist(
        x[:-1], bins=x, weights=y,
        color='#1f78b4', alpha=0.8,
        orientation='horizontal',
        label=r'$-1\sigma$',
    )
    y = data["poi_paramup"].values
    ax[1].hist(
        x[:-1], bins=x, weights=y,
        color='#e31a1c', alpha=0.8,
        orientation='horizontal',
        label=r'$+1\sigma$',
    )
    xmax = np.max(np.abs(ax[1].get_xlim()))
    ax[1].set_xlim(-1.1*xmax, 1.1*xmax)
    ax[1].set_ylim(0, len(y))
    ax[1].axvline(0, lw=1, color='gray', alpha=0.8)

    y = data["param_value"].values
    yerr = (
        -1*data["param_merrdown"].values,
        data["param_merrup"].values,
    )
    ax[0].errorbar(
        y, (x[:-1]+x[1:])/2., xerr=yerr,
        fmt='o', color='black',
        ms=4, capsize=4,
    )
    xmax = data.eval("param_value+param_merrup").max()
    xmax = max(xmax, data.eval("-(param_value+param_merrdown)").max())
    xmax = int(xmax)+1
    ax[0].set_xlim(-xmax, xmax)
    for pos in range(xmax):
        ax[0].axvline(pos, lw=1, color='gray', alpha=0.8)
        ax[0].axvline(-pos, lw=1, color='gray', alpha=0.8)
    ax[0].set_ylim(0, len(y))
    ax[0].set_xticks(np.arange(-(xmax-1), (xmax-1)+0.1, 1.))
    ax[0].set_yticks((x[:-1]+x[1:])/2.)
    labels = [
        converter.get(l, l.replace("_", " "))
        for l in data.index.get_level_values("param").values
    ]
    ax[0].set_yticklabels(labels)
    ax[0].set_xlabel(r'$\theta$')
    ax[1].set_xlabel(r'$\Delta\hat{r}$')
    ax[1].legend(fancybox=True, edgecolor='#d9d9d9')
    return fig, ax

def nllscan(
    x, y, ax=None, marker_kw={}, spline_kw={}, splrep_kw={}, splev_kw={},
    opt_kw={}, root_kw={}, line_kw={}, text_kw={}, nsigs=[1],
    bestfit_guess=[0.], left_bracket=(-np.inf, 0), right_bracket=(0, np.inf),
):
    """
    Helper function to plot a -2*Delta(log(L)) scan from a pd.DataFrame with
    two columns: x variable and y variable (which should hold the
    -2*Delta(log(L)) values.

    Parameters
    ----------
    x : np.ndarray-like
        The input x variable.

    y : np.ndarray-like
        The input y variable. Should hold values of -2*Delta(log(L))

    ax : matplotlib.axes, optional (default=None)
        The axis to draw on.

    marker_kw : dict-like, optional (default={})
        kwargs to pass to ax.plot. Updates a dict with:
        dict(marker='o', ms=2, lw=0., label='Scan', color='#1f78bf')

    spline_kw : dict-like, optional (default={})
        kwargs to pass to ax.plot. Updates a dict with:
        dict(lw=1., label='Spline', color='#e31a1c')

    splrep_kw: dict-like, optional (default={})
        kwargs to pass to scipy.interpolate.splrep. Updates a dict with:
        dict(s=0)

    splev_kw: dict-like, optional (default={})
        kwargs to pass to scipy.interpolate.splev. Updates a dict with:
        dict(der=0)

    opt_kw: dict-like, optional (default={})
        kwargs to pass to scipy.optimize.optimize. Updates a dict with:
        dict(der=0)

    root_kw: dict-like, optional (default={})
        kwargs to pass to scipy.optimize.root_scalar. Updates a dict with:
        dict(method='brentq')

    line_kw: dict-like, optional (default={})
        kwargs to pass to axes.ax?line. Updates a dict with:
        dict(lw=1, ls='--', color='gray')

    text_kw: dict-like, optional (default={})
        kwargs to pass to axes.text. Updates a dict with:
        ict(ha='left', va='bottom', color='gray')

    nsigs : list of floats, optional (default=[1])
        List of number of sigmas to draw on the final plot

    bestfit_guess : list of floats, options (default=[0.])
        Best fit guess of the minimum for scipy.optimize

    left_bracket : tuple of floats, options (default=(-np.inf, 0))
        Guess for left root bracket.

    right_bracket : tuple of floats, options (default=(-np.inf, 0))
        Guess for right root bracket.

    Return
    ------
    pd.DataFrame with columns: nsig and x values
    """
    outdata = []
    if ax is None:
        fig, ax = plt.subplots()

    kw = dict(marker='o', ms=2, lw=0., label='Scan', color='#1f78bf')
    kw.update(marker_kw)
    ax.plot(x, y, **kw)

    # spline
    kw = dict(s=0)
    kw.update(splrep_kw)
    tck = interp.splrep(x, y, **kw)

    kw = dict(der=0)
    kw.update(splev_kw)
    kw["tck"] = tck
    func = functools.partial(interp.splev, **kw)

    xfine = np.linspace(x.min(), x.max(), 201)
    kw = dict(lw=1., label='Spline', color='#e31a1c')
    kw.update(spline_kw)
    ax.plot(xfine, func(xfine), **kw)

    kw = dict(method='L-BFGS-B')
    kw.update(opt_kw)
    bestfit = opt.minimize(func, bestfit_guess, **kw)
    outdata.append({"nsig": 0., "xval": bestfit.x[0]})

    for nsig in nsigs:
        kw = dict(method='brentq')
        kw.update(root_kw)
        kw["bracket"] = left_bracket
        left = opt.root_scalar(lambda x: func(x)-nsig**2, **kw)
        outdata.append({"nsig": nsig, "xval": left.root})

        kw = dict(method='brentq')
        kw.update(root_kw)
        kw["bracket"] = right_bracket
        right = opt.root_scalar(lambda x: func(x)-nsig**2, **kw)
        outdata.append({"nsig": -nsig, "xval": right.root})

        kw = dict(lw=1, ls='--', color='gray')
        kw.update(line_kw)
        ax.axvline(left.root, **kw)
        ax.axvline(right.root, **kw)
        ax.axhline(nsig**2, **kw)

        pos = ax.transData.inverted().transform(
            ax.transAxes.transform((0.025, 1))
        )
        kw = dict(ha='left', va='bottom', color='gray')
        kw.update(text_kw)
        ax.text(pos[0], nsig**2, f'${nsig}\\sigma$', **kw)

    return pd.DataFrame(outdata)
